library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.numeric_std.ALL;

library unisim;
use unisim.vcomponents.all;

use work.IOB_Config.ALL;
use work.VT52_cfg.ALL;

entity iob is
	port
	(clk : in std_logic;	-- 28.63636 MHz

	-- CPU bus interface
	cpu_ioclr_n : in std_logic;
	dx : inout std_logic_vector(0 to 11);
	clk_write_n : in std_logic;
	cpu_read_n : in std_logic;
	cpu_sr_read_n : in std_logic;
	cpu_sr_write_n : in std_logic;
	clk_lxdar_n : in std_logic;
	cpu_c0_n : out std_logic;
	cpu_c1_n : out std_logic;
	cpu_skip_n : out std_logic;
	cpu_intreq_n : out std_logic;
	cpu_intgrnt_n : in std_logic;
	drive_dx_l: out std_logic;

	-- serial ports
	txd : out std_logic_vector(0 to NUMUARTS-1);
	rxd : in std_logic_vector(0 to NUMUARTS-1);

	-- LPT port
	lpt_ack: in std_logic;
	lpt_busy_n: in std_logic;
	lpt_paper_end_n: in std_logic;
	lpt_select_in_n: in std_logic;
	lpt_error: in std_logic;
	lpt_strobe: out std_logic;
	lpt_ddir: out std_logic;
	lpt_data: inout std_logic_vector(7 downto 0);
	lpt_init: out std_logic;
		
	-- Keyboard
	kb_clk: inout std_logic;
	kb_data: inout std_logic;

	-- VGA
	vga_RGB: out RGB;
	vga_HS: out std_logic;
	vga_VS: out std_logic;

	-- bits for custom applications
	iobits: inout std_logic_vector(0 to 35);

	-- special purpose
	reprogram: out std_logic;	-- drive low to restart FPGA configuration
	cpreq_n: out std_logic);
end iob;

architecture RTL of iob is

	-- PosEdge: in this implementation, most processes are
	-- synchronous to clk, so this component is used to sample
	-- slower signals, and generates a 1-clock pulse on their rising edge
	component PosEdge
		port
		  (reset: in boolean;
			clk: in std_logic;
			inp: in std_logic;          
			outp: out std_logic);
	end component;

	component KL8JA is
		generic
			(print_addr: DevID;
			keyb_addr: DevID);
		port
			( -- clocks
			clk: in std_logic;
			brg: in unsigned(7 downto 0);

			-- CPU bus interface
			reset : in boolean;
			IOTact: in boolean;
			IOTdev: in DevID;
			IOTcmd: in DevCmd;
			dx : inout std_logic_vector(0 to 11);
			drive_dx_l: out std_logic;
			clk_write_n : in std_logic;
			clk_lxdar_n : in std_logic;
			IOTread : in boolean;
			cpu_c0_n : out std_logic;
			cpu_c1_n : out std_logic;
			cpu_skip_n : out std_logic;

			-- communication
			IRQ: out std_logic;

			-- serial port
			txd : out std_logic;
			rxd : in std_logic
			);
	end component;

	component LC8E is
		generic
			(print_addr: DevID;
			 keyb_addr: DevID);
		port
			(reset: in boolean;							-- true when reset (IOCLR)
			 IOTact: in boolean;							-- true when in IOT (LXDAR)
			 IOTdev: in DevID;							-- latched IOT device code
			 IOTcmd: in DevCmd;							-- latched IOT command

			 clk_write_n : in std_logic;
			 dx : in std_logic_vector(7 downto 0);
			 drive_dx_l: out std_logic;
			 cpu_skip_n : out std_logic;
			 pIRQ: out std_logic;

			 -- LPT port
			 lpt_ack: in std_logic;
			 lpt_busy_n: in std_logic;
			 lpt_paper_end_n: in std_logic;
			 lpt_select_in_n: in std_logic;
			 lpt_error: in std_logic;
			 lpt_strobe: buffer std_logic;
			 lpt_ddir: out std_logic;
			 lpt_data: inout std_logic_vector(7 downto 0);
			 lpt_init: out std_logic);
	end component;

	component VT52 is
		generic
			(print_addr: DevID;
			 keyb_addr: DevID);
		port
			( -- clocks
			clk : in std_logic;
			kb_clk_src: in std_logic;

			-- CPU bus interface
			reset : in boolean;
			IOTact: in boolean;
			IOTdev: in DevID;
			IOTcmd: in DevCmd;

			clk_write_n : in std_logic;
		 	clk_lxdar_n : in std_logic;
			IOTread : in boolean;
			dx : inout std_logic_vector(0 to 11);
			drive_dx_l: out std_logic;
			cpu_c0_n : out std_logic;
			cpu_c1_n : out std_logic;
			cpu_skip_n : out std_logic;
			IRQ: out std_logic;

			-- Keyboard
			kb_clk: inout std_logic;
			kb_data: inout std_logic;

			-- VGA
			vga_RGB: out std_logic_vector(0 to 2);
			vga_HS: out std_logic;
			vga_VS: out std_logic);
	end component;

	component Div3 is
	    port
		 	(I: in std_logic;
			 O: out std_logic);
	end component;

-- registers & signals shared between all devices

signal cpu_ioclr: std_logic;
signal reset: boolean;							-- true when reset (IOCLR)
signal IOTact: boolean;							-- true when in IOT (LXDAR)
signal IOTdev: DevID;							-- latched IOT device code
signal IOTcmd: DevCmd;							-- latched IOT command
signal IOTread: boolean;						-- true during IOT read cycle

signal dx_byte: std_logic_vector(7 downto 0); -- dx "right way round"

signal clkfb, clkm2x, kb_clk_src: std_logic;
signal prescale: unsigned(6 downto 0);
signal brg: unsigned(7 downto 0);			-- baud rate generator

signal uIRQ: unsigned(0 to NUMUARTS-1);	-- masked flags from the serial ports
signal pIRQ: std_logic;							-- masked flag from the printer port
signal vIRQ: std_logic;							-- masked flag from VT52

-- dummy switch register registers & signals 

signal switchreg: std_logic_vector(0 to 11);
signal sr_write: std_logic;

begin

-----------------------------------------------------------------------------------------
-- utility and shared logic
-----------------------------------------------------------------------------------------

reprogram <= '1';		-- not implemented yet
cpreq_n <= 'Z';

dx_byte(7 downto 0) <= dx(4 to 11);

-- common IOT decoding

cpu_ioclr <= not cpu_ioclr_n;
reset <= cpu_ioclr_n = '0';

-- latch IOT device and command at start of LXDAR
process (clk_lxdar_n)
begin
	if falling_edge(clk_lxdar_n) then
		IOTdev <= unsigned(dx(3 to 8));
		IOTcmd <= unsigned(dx(9 to 11));
	end if;
end process;

IOTact <= (clk_lxdar_n = '0');

-- here we stretch the read signal from when it is active
-- until the end of LXDAR; IM6120 wants IOT read data
-- at end of LXDAR, not READ
process (reset, clk, IOTact, cpu_read_n)
begin
	if reset or not IOTact then
		IOTread <= false;
	elsif rising_edge(clk) then
		if cpu_read_n = '0' then
			IOTread <= true;
		end if;
	end if;
end process;

---- bring CPU IRQ low when printer or serial ports request it
cpu_intreq_n <= '0' when pIRQ = '1' or vIRQ = '1' or (uIRQ /= 0) else 'Z';

-- baud rate generator for uarts

-- to simplify the circuit, we want to use one oscillator for both
-- the video and the serial clocks.  Unfortunately there's nothing
-- close to the VGA text mode frequency that divides evenly to standard
-- baud rates.  So the VGA clock is chosen to be reasonably close to
-- the right dot clock _and_ doesn't have too much error when divided
-- to the baud clock (16x the baud rate).
--
-- the second wrinkle is that we use a special Xilinx feature to double
-- the clock rate, which therefore reduces the error in half when it's
-- divided down.  This is done by instantiating some library objects that
-- represent features built into the FPGA.

clk1: clkdll port map (RST => cpu_ioclr, CLKIN => clk,
							  CLKFB => clkfb, CLK2x => clkm2x);

fbnet: BUFG port map (I => clkm2x, O => clkfb);

brg(0) <= prescale(6);

process (reset, clkm2x)
begin
	if reset then
		brg(7 downto 1) <= (others => '0');
		prescale <= (others => '0');
	elsif rising_edge(clkm2x) then
		if to_integer(prescale) = 93 then
			prescale <= (others => '0');
			brg(7 downto 1) <= brg(7 downto 1) + 1;
		else
			prescale <= prescale + 1;
		end if;
	end if;
end process;

-- clock generator for PS/2 keyboard output - needs to be between 10 and 16 KHz,
-- so we divide the 2400*16 baud clock by 3 for 12.8 KHz

kbcgen: Div3 port map (I => brg(4), O => kb_clk_src);

-----------------------------------------------------------------------------------------
-- switch register dummy implementation - just a latch
-----------------------------------------------------------------------------------------

process (cpu_sr_read_n, dx, switchreg)
begin
	if cpu_sr_read_n = '0' then
		dx <= switchreg;
		drive_dx_l <= '0';
	else
		dx <= (others => 'Z');
		drive_dx_l <= 'Z';
	end if;
end process;

sr_w_detect: PosEdge port map (reset, clk, cpu_sr_write_n, sr_write);

process (clk, sr_write)
begin
	if rising_edge(clk) then
		if sr_write = '1' then
			switchreg <= dx;
		end if;
	end if;
end process;

-----------------------------------------------------------------------------------------
-- KL8JA serial ports
-----------------------------------------------------------------------------------------

KL8JAs: for idx in 0 to NUMUARTS-1 generate
begin

	sp: KL8JA
		generic map
		(
		print_addr => UARTOBASE(idx),
		keyb_addr => UARTIBASE(idx)
		)
		port map
			(clk => clk, reset => reset,
			 brg => brg, -- note: LSB of brg won't have 50% duty cycle
			 IOTact => IOTact, IOTdev => IOTdev, IOTread => IOTread,
			 IOTcmd => IOTcmd, dx => dx, clk_write_n => clk_write_n, clk_lxdar_n => clk_lxdar_n,
			 cpu_c0_n => cpu_c0_n, cpu_c1_n => cpu_c1_n, cpu_skip_n => cpu_skip_n,
			 IRQ => uIRQ(idx), txd => txd(idx), rxd => rxd(idx));

end generate;

-----------------------------------------------------------------------------------------
-- LC8E printer interface with Centronics output
-----------------------------------------------------------------------------------------

parpt_if: if PARPTIBASE /= O"00" generate
signal lpt_strobe_i: std_logic;
begin

lpt_strobe <= lpt_strobe_i;

pp: LC8E
	generic map
		(print_addr => PARPTOBASE, keyb_addr => PARPTIBASE)
	port map
		(reset => reset, IOTact => IOTact, clk_write_n => clk_write_n,
		 IOTdev => IOTdev, IOTcmd => IOTcmd, dx => dx_byte,
		 cpu_skip_n => cpu_skip_n, pIRQ => pIRQ,
		 lpt_ack => lpt_ack, lpt_busy_n => lpt_busy_n,
		 lpt_paper_end_n => lpt_paper_end_n,
		 lpt_select_in_n => lpt_select_in_n, lpt_error => lpt_error,
		 lpt_strobe => lpt_strobe_i, lpt_ddir => lpt_ddir,
		 lpt_data => lpt_data, lpt_init => lpt_init);

end generate;

-----------------------------------------------------------------------------------------
-- VT52 via KL8E interface
-----------------------------------------------------------------------------------------

vt: VT52
	generic map
		(print_addr => VT52OBASE, keyb_addr => VT52IBASE)
	port map
		(clk => clk,	kb_clk_src => kb_clk_src, reset => reset,
		 IOTact => IOTact, IOTdev => IOTdev,	IOTcmd => IOTcmd,
		 clk_write_n => clk_write_n, clk_lxdar_n => clk_lxdar_n,
		 IOTread => IOTread, dx => dx,
		 cpu_c0_n => cpu_c0_n, cpu_c1_n => cpu_c1_n,
		 cpu_skip_n => cpu_skip_n, IRQ => vIRQ,
		 kb_clk => kb_clk, kb_data => kb_data,
		 vga_RGB => vga_RGB,
		 vga_HS => vga_HS,	vga_VS => vga_VS);

-----------------------------------------------------------------------------------------
-- Port bits IOT
-- for now, just implement programmable I/O ports, 3x 12 bit

iobits_if: if PIOBASE /= O"00" generate

	signal bits_bank: unsigned(0 to 1);
	signal bits_sel: boolean;
	signal bits_dir: std_logic_vector(iobits'range);
	signal bits_out: std_logic_vector(iobits'range);

begin

	process (IOTact, IOTdev)
	begin
		bits_sel <= IOTact and (IOTdev = PIOBASE);
	end process;

	process (bits_out, bits_dir)
	begin
		for i in iobits'range loop
			if bits_dir(i) = '0' then
				iobits(i) <= 'Z';
			else
				iobits(i) <= bits_out(i);
			end if;
		end loop;
	end process;

	-- reading
	process (IOTread, bits_sel, bits_bank, iobits)
	begin
		if bits_sel and IOTread then
			case bits_bank is
				when "01" => dx <= iobits(12 to 23);
				when "10" => dx <= iobits(24 to 35);
				when others => dx <= iobits(0 to 11);
			end case;
			drive_dx_l <= '0';
		else
			dx <= (others => 'Z');
			drive_dx_l <= 'Z';
		end if;
	end process;

	-- writing
	process (reset, bits_sel, clk_write_n, dx)
	begin
		if reset then
			bits_dir <= (others => '0');	-- inputs on reset
		elsif rising_edge(clk_write_n) then
			if bits_sel then
				if IOTcmd = PIOSETBITS then -- write bits
					case bits_bank is
						when "01" => bits_out(12 to 23) <= dx;
						when "10" => bits_out(24 to 35) <= dx;
						when others => bits_out(0 to 11) <= dx;
					end case;
				elsif IOTcmd = PIOSETBANK then	-- select bank
					bits_bank <= unsigned(dx(10 to 11));
				elsif IOTcmd = PIOSETDIR then -- set direction
					case bits_bank is
 						when "01" => bits_dir(12 to 23) <= dx;
 						when "10" => bits_dir(24 to 35) <= dx;
 						when others => bits_dir(0 to 11) <= dx;
 					end case;
				end if;
			end if;
		end if;
	end process;

end generate; -- if PIOBASE /= O"00"

end RTL;
